<?php

namespace Drupal\cloudflare_sitemap;

use Cloudflare\API\Adapter\Guzzle;
use Cloudflare\API\Auth\APIToken;
use Cloudflare\API\Endpoints\Zones;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\State\StateInterface;
use GuzzleHttp\ClientInterface;
use Psr\Log\LoggerInterface;

/**
 * Class implements service cloudflare_sitemap.cache_purger.
 */
class CachePurger {

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * A logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The config for Cloudflare API.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * Constructs a CachePurger object.
   *
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The HTTP client.
   * @param \Psr\Log\LoggerInterface $logger
   *   Logger channel.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   Config factory.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   */
  public function __construct(ClientInterface $http_client, LoggerInterface $logger, ConfigFactoryInterface $config_factory, StateInterface $state) {
    $this->httpClient = $http_client;
    $this->logger = $logger;
    $this->config = $config_factory->get('cloudflare.settings');
    $this->state = $state;
  }

  /**
   * Ask Cloudflare API to purge cache for specified URLs.
   *
   * @param array $urls
   *   Array of URLs from any supported domains.
   *
   * @return bool
   *   TRUE if Cloudflare accepted the request.
   */
  public function purgeFromCloudflare(array $urls = []) {
    // No need to purge cache if website in the maintenance mode.
    if ($this->state->get('system.maintenance_mode')) {
      $this->logger->debug('Cache purge is skipped, because maintenance mode is enabled.');
      return TRUE;
    }

    $api_token = $this->config->get('api_token');
    $zone_id = $this->config->get('zone_id');

    // We save all URLs purged in one request to avoid unnecessary duplicated
    // API calls to Cloudflare.
    $already_flushed_urls = &drupal_static('cloudflare_already_purged_urls', []);

    if (empty($urls) || empty($api_token) || empty($zone_id)) {
      // It's fine to return TRUE if it's DRY RUN mode (dev environments).
      return TRUE;
    }

    $not_purged_urls = array_diff($urls, $already_flushed_urls);
    if (!$not_purged_urls) {
      return TRUE;
    }

    // Prepare two arrays: one for Cloudflare API (with optional origin header)
    // and another one with URLs as strings for storing in static cache.
    $purge_urls = [];
    $purge_urls_as_strings = [];
    foreach ($urls as $url) {
      if (empty($url)) {
        continue;
      }
      $purge_urls_as_strings[$url] = $url;
      $url_with_origin = explode("|With-Origin:", $url);
      if (!empty($url_with_origin) && count($url_with_origin) === 2) {
        $purge_urls[$url] = [
          'url' => $url_with_origin[0],
          'headers' => [
            // Origin header is needed to flush XHR requests cache.
            'Origin' => $url_with_origin[1],
          ],
        ];
      }
      else {
        $purge_urls[$url] = $url;
      }
    }
    // Cloudflare API accepts up to 30 URLs per API call.
    $purge_urls = array_chunk($purge_urls, 30);
    $purge_urls_as_strings = array_chunk($purge_urls_as_strings, 30);

    try {
      // Init Cloudflare API.
      $key = new APIToken($api_token);
      $adapter = new Guzzle($key);
      $zonesApi = new Zones($adapter);

      $status = FALSE;

      foreach ($purge_urls as $i => $urls_chunk) {
        $status = $zonesApi->cachePurge($zone_id[0], $urls_chunk);

        if ($status) {
          // Add flushed URLs in static variable. Note that we use plain strings
          // to simplify comparing of arrays in $this->purgeMultiple.
          $already_flushed_urls = array_merge($already_flushed_urls, $purge_urls_as_strings[$i]);

          $this->logger->info('Purged URLs from Cloudflare (part @part): <pre>@urls</pre>', [
            '@urls' => print_r($urls_chunk, TRUE),
            '@part' => $i + 1,
          ]);
        }
        else {
          // PHP SDK may just return FALSE without specifying an error.
          $this->logger->error('Error purging URLs from Cloudflare: <pre>@urls</pre>.', ['@urls' => print_r($urls_chunk, TRUE)]);

          // Do not continue if an error occurred.
          break;
        }
      }

      return $status;

    }
    catch (\Exception $exception) {
      $this->logger->error('Error purging URLs from Cloudflare. Error %error. URLs: <pre>@urls</pre>.',
        [
          '%error' => $exception->getMessage(),
          '@urls' => print_r($urls, TRUE),
        ]
      );
      return FALSE;
    }
    catch (\Throwable $exception) {
      $this->logger->error('Error purging URLs from Cloudflare. Error %error. URLs: <pre>@urls</pre>.',
        [
          '%error' => $exception->getMessage(),
          '@urls' => print_r($urls, TRUE),
        ]
      );
      return FALSE;
    }

  }

  /**
   * Purges URLs from the external sitemap URL.
   *
   * @param string $sitemap_url
   *   Sitemap URL.
   *
   * @return bool
   *   TRUE if Cloudflare cache was purged. Otherwise, FALSE.
   */
  public function purgeSitemap($sitemap_url) {
    // No need to purge cache if website in the maintenance mode.
    if ($this->state->get('system.maintenance_mode')) {
      $this->logger->debug('Sitemap cache purge is skipped, because maintenance mode is enabled.');
      return TRUE;
    }

    try {
      $urls = [];
      $file = file_get_contents($sitemap_url);
      $sitemap = new \SimpleXMLElement($file);
      foreach ($sitemap->url as $url_list) {
        $urls[] = (string) $url_list->loc;
      }

      return $this->purgeFromCloudflare($urls);
    }
    catch (\Exception $exception) {
      $this->logger->error('Error purging sitemap %sitemap from Cloudflare. Error %error.', ['%sitemap' => $sitemap_url, '%error' => $exception->getMessage()]);
      return FALSE;
    }
  }

}
